# -*- coding: utf-8 -*-
#
# This file is part of the RCUSCC project
#
#
#
# Distributed under the terms of the APACHE license.
# See LICENSE.txt for more info.

""" RCU-SCC Device Server for LOFAR2.0

"""

# PyTango imports
import tango
from tango import DebugIt
from tango.server import run
from tango.server import Device
from tango.server import attribute, command
from tango.server import device_property
from tango import AttrQuality, DispLevel, DevState
from tango import AttrWriteType, PipeWriteType
# Additional import
# PROTECTED REGION ID(RCUSCC.additionnal_import) ENABLED START #
import sys
import opcua
import traceback
import numpy
import time
from threading import Thread

from functools import wraps
import socket

def only_when_on(func):
    """
      Wrapper to return None when the device isn't in ON state.

      If in ON state, calls & returns the wrapped function.
    """

    @wraps(func)
    def when_on_wrapper(self, *args, **kwargs):
        if self.get_state() != DevState.ON:
            return None

        return func(self, *args, **kwargs)

    return when_on_wrapper

def fault_on_opcua_error(func):
    """
      Wrapper to catch exceptions generated by the OPC-UA connection.

      Sets the device in a FAULT state if that occurs.
    """

    @wraps(func)
    def opcua_error_wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            self.error_stream("Communication with the server %s failed.  Trace: %s", self.client.server_url.geturl(), traceback.format_exc())
            self.Fault()
            return None

    return opcua_error_wrapper
# PROTECTED REGION END #    //  RCUSCC.additionnal_import

__all__ = ["RCUSCC", "main"]

class OPCUAConnection(Thread):
    """
      Run a connector function in the background, until it succeeds.
    """

    def __init__(self, client, init_func, fault_func, streams, try_interval=2):
        super().__init__(daemon=True)

        self.client = client
        self.init_func = init_func
        self.fault_func = fault_func
        self.try_interval = try_interval
        self.streams = streams
        self.stopping = False

        self.start()

    def _servername(self):
        return self.client.server_url.geturl()

    def try_connect(self):
        try:
            self.streams.debug_stream("Connecting to server %s", self._servername())
            self.client.connect()
            self.streams.debug_stream("Connected to server. Initialising.")

            self.init_func()

            return True
        except socket.error as e:
            self.streams.error_stream("Could not connect to server %s: %s", self._servername(), e)

            # signal that we're disconnected
            self.fault_func()
            return False

    def try_disconnect(self):
        try:
            self.client.disconnect()
        except Exception as e:
            self.streams.error_stream("Disconnect from OPC-UA server %s failed: %s", self._servername(), e)

    def run(self):
        while not self.stopping:
            # keep trying to connect
            while not self.stopping and not self.try_connect(): 
               time.sleep(self.try_interval)

            # keep checking if the connection is still alive
            try:
                while not self.stopping:
                    self.client.send_hello()
                    time.sleep(self.try_interval)
            except Exception as e:
                self.streams.error_stream("Lost connection to server %s: %s", self._servername(), e)

                # technically, we may not have dropped the connection, but encounter a different error. so explicitly disconnect.
                self.try_disconnect()

                # signal that we're disconnected
                self.fault_func()

    def stop(self):
        """
          Stop connecting & disconnect. Can take a few seconds for the timeouts to hit.
        """

        self.stopping = True
        self.join()

        self.try_disconnect()

class RCUSCC(Device):
    """

    **Properties:**

    - Device Property
        OPC_Server_Name
            - Type:'DevString'
        OPC_Server_Port
            - Type:'DevULong'
        OPC_Time_Out
            - Type:'DevDouble'
    """
    # PROTECTED REGION ID(RCUSCC.class_variable) ENABLED START #
    client = 0
    name_space_index = 0
    obj = 0
    # PROTECTED REGION END #    //  RCUSCC.class_variable

    # -----------------
    # Device Properties
    # -----------------

    OPC_Server_Name = device_property(
        dtype='DevString',
        mandatory=True
    )

    OPC_Server_Port = device_property(
        dtype='DevULong',
        mandatory=True
    )

    OPC_Time_Out = device_property(
        dtype='DevDouble',
        mandatory=True
    )

    # ----------
    # Attributes
    # ----------

    Attenuator_R = attribute(
        dtype=(('DevLong64',),),
        max_dim_x=32, max_dim_y=3,
    )

    Attenuator_RW = attribute(
        dtype=(('DevLong64',),),
        max_dim_x=32, max_dim_y=3,
        access=AttrWriteType.READ_WRITE,
    )

    Band_R = attribute(
        dtype=(('DevLong64',),),
        max_dim_x=32, max_dim_y=3,
    )

    Band_RW = attribute(
        dtype=(('DevLong64',),),
        max_dim_x=32, max_dim_y=3,
        access=AttrWriteType.READ_WRITE,
    )

    CLK_PLL_locked_R = attribute(
        dtype='DevLong64',
    )

#    Dither_Frequency_R = attribute(
#        dtype=(('DevDouble',),),
#        max_dim_x=32, max_dim_y=2,
#    )
#
#    Dither_Frequency_RW = attribute(
#        dtype=(('DevDouble',),),
#        max_dim_x=32, max_dim_y=2,
#        access=AttrWriteType.READ_WRITE,
#    )

    LED_R = attribute(
        dtype=(('DevLong64',),),
        max_dim_x=32, max_dim_y=3,
    )

    LED_RW = attribute(
        dtype=(('DevLong64',),),
        max_dim_x=32, max_dim_y=3,
        access=AttrWriteType.READ_WRITE,
    )

    Pwr_dig_R = attribute(
        dtype=('DevLong64',),
        max_dim_x=32,
    )

    Temperature_R = attribute(
        dtype=('DevDouble',),
        max_dim_x=32,
    )


    # ---------------
    # General methods
    # ---------------

    def get_pcc_node(self, node):
        try:
            return self.pcc_node.get_child(["{}:{}".format(self.name_space_index, node)])
        except opcua.ua.uaerrors._auto.BadNoMatch:
            self.error_stream("Could not find PCC node %s, using a dummy", node)

            class DummyNode:
                def get_value(self):
                    pass

                def set_value(self, val):
                    pass

            return DummyNode()


    def _init_opcua(self):
        try:
            self.name_space_index = self.client.get_namespace_index("http://lofar.eu")
        except Exception as e:
            self.warn_stream("Cannot determine the OPC-UA name space index.  Will try and use the default = 2.")
            self.name_space_index = 2

        self.obj_node = self.client.get_objects_node()
        self.pcc_node = self.obj_node.get_child(["{}:PCC".format(self.name_space_index)])

        self.debug_stream("Mapping OPC-UA MP/CP to attributes...")

        # 2020-11-27, thomas
        # TODO
        # Modify as soon as we have a real multi-dimensional array
        self.attribute_mapping["Attenuator_R"] = (self.get_pcc_node("RCUs_Attenuator1_R"), self.get_pcc_node("RCUs_Attenuator2_R"), self.get_pcc_node("RCUs_Attenuator3_R"))

        # 2020-11-27, thomas
        # TODO
        # Modify as soon as we have a real multi-dimensional array
        self.attribute_mapping["Attenuator_RW"] = (self.get_pcc_node("RCUs_Attenuator1_RW"), self.get_pcc_node("RCUs_Attenuator2_RW"), self.get_pcc_node("RCUs_Attenuator3_RW"))

        # 2020-11-27, thomas
        # TODO
        # Modify as soon as we have a real multi-dimensional array
        self.attribute_mapping["Band_R"] = (self.get_pcc_node("RCUs_Band1_R"), self.get_pcc_node("RCUs_Band2_R"), self.get_pcc_node("RCUs_Band3_R"))

        # 2020-11-27, thomas
        # TODO
        # Modify as soon as we have a real multi-dimensional array
        self.attribute_mapping["Band_RW"] = (self.get_pcc_node("RCUs_Band1_RW"), self.get_pcc_node("RCUs_Band2_RW"), self.get_pcc_node("RCUs_Band3_RW"))

        self.attribute_mapping["CLK_PLL_locked_R"] = self.get_pcc_node("CLK_PLL_locked_R")

#            self.attribute_mapping["Dither_Frequency_R"] = self.get_pcc_node("RCUs_Dither_Frequency_R")

#            self.attribute_mapping["Dither_Frequency_RW"] = self.get_pcc_node("RCUs_Dither_Frequency_RW")

        self.attribute_mapping["LED_R"] = self.get_pcc_node("RCUs_LED0_R")

        self.attribute_mapping["LED_RW"] = self.get_pcc_node("RCUs_LED0_RW")

        self.attribute_mapping["Pwr_dig_R"] = self.get_pcc_node("RCUs_Pwr_dig_R")

        self.attribute_mapping["Temperature_R"] = self.get_pcc_node("RCUs_Temperature_R")

#            self.attribute_mapping["RCU_off"] = self.get_pcc_node("RCUs_RCU_off")

#            self.attribute_mapping["RCU_on"] = self.get_pcc_node("RCUs_RCU_on")


        self.debug_stream("Mapping OPC-UA MP/CP to attributes done.")

        # we have connected, so are on-line.
        self.On()


    def init_device(self):
        """Initialises the attributes and properties of the RCUSCC."""
        Device.init_device(self)

        # PROTECTED REGION ID(RCUSCC.init_device) ENABLED START #
        self.set_state(DevState.INIT)

        # Init the dict that contains attribute to OPC-UA MP/CP mappings.
        self.attribute_mapping = {}

        # Set default values in the RW/R attributes and add them to
        # the mapping.
        self._Attenuator_R = ((0,),)
        self.attribute_mapping["Attenuator_R"] = {}

        self._Attenuator_RW = ((0,),)
        self.attribute_mapping["Attenuator_RW"] = {}

        self._CLK_PLL_locked_R = 0
        self.attribute_mapping["CLK_PLL_locked_R"] = {}

        self._Band_R = ((0,),)
        self.attribute_mapping["Band_R"] = {}

        self._Band_RW = ((0,),)
        self.attribute_mapping["Band_RW"] = {}

        self._Dither_Frequency_R = ((0.0,),)
        self.attribute_mapping["Dither_Frequency_R"] = {}

        self._Dither_Frequency_RW = ((0.0,),)
        self.attribute_mapping["Dither_Frequency_RW"] = {}

        self._LED_R = ((0,),)
        self.attribute_mapping["LED_R"] = {}

        self._LED0_RW = ((0,),)
        self.attribute_mapping["LED_RW"] = {}

        self._Pwr_dig_R = (0,)
        self.attribute_mapping["Pwr_dig_R"] = {}

        self._Temperature_R = (0.0,)
        self.attribute_mapping["Temperature_R"] = {}

        self.client = opcua.Client("opc.tcp://{}:{}/".format(self.OPC_Server_Name, self.OPC_Server_Port), self.OPC_Time_Out) # timeout in seconds

        # Connect to OPC-UA -- will set ON state on success
        self.opcua_connection = OPCUAConnection(self.client, self._init_opcua, self.Fault, self)

        # PROTECTED REGION END #    //  RCUSCC.init_device

    def always_executed_hook(self):
        """Method always executed before any TANGO command is executed."""
        # PROTECTED REGION ID(RCUSCC.always_executed_hook) ENABLED START #
        # PROTECTED REGION END #    //  RCUSCC.always_executed_hook

    def _disconnect(self):
        try:
            if self.client is not None:
                self.client.disconnect()
        except Exception as e:
            self.warn_stream("Disconnect from OPC-UA server %s:%d failed.  Trace: %s" % (self.OPC_Server_Name, self.OPC_Server_Port, traceback.format_exc()))
        finally:
            self.client = None

    def delete_device(self):
        """Hook to delete resources allocated in init_device.

        This method allows for any memory or other resources allocated in the
        init_device method to be released.  This method is called by the device
        destructor and by the device Init command.
        """
        # PROTECTED REGION ID(RCUSCC.delete_device) ENABLED START #
        self.debug_stream("Shutting down...")

        self.Off()
        self.debug_stream("Shut down.  Good bye.")
        # PROTECTED REGION END #    //  RCUSCC.delete_device
    # ------------------
    # Attributes methods
    # ------------------

    @only_when_on
    @fault_on_opcua_error
    def read_Attenuator_R(self):
        # PROTECTED REGION ID(RCUSCC.Attenuator_R_read) ENABLED START #
        """Return the Attenuator_R attribute."""
        self._Attenuator_R = numpy.array([self.attribute_mapping["Attenuator_R"][0].get_value(), self.attribute_mapping["Attenuator_R"][1].get_value(), self.attribute_mapping["Attenuator_R"][2].get_value()])
        return self._Attenuator_R
        # PROTECTED REGION END #    //  RCUSCC.Attenuator_R_read

    @only_when_on
    @fault_on_opcua_error
    def read_Attenuator_RW(self):
        # PROTECTED REGION ID(RCUSCC.Attenuator_RW_read) ENABLED START #
        """Return the Attenuator_R attribute."""
        self._Attenuator_RW = numpy.array([self.attribute_mapping["Attenuator_RW"][0].get_value(), self.attribute_mapping["Attenuator_RW"][1].get_value(), self.attribute_mapping["Attenuator_RW"][2].get_value()])
        return self._Attenuator_RW
        # PROTECTED REGION END #    //  RCUSCC.Attenuator_RW_read

    @only_when_on
    @fault_on_opcua_error
    def write_Attenuator_RW(self, value):
        # PROTECTED REGION ID(RCUSCC.Attenuator_RW_write) ENABLED START #
        """Set the Attenuator_RW attribute."""
        self.attribute_mapping["Attenuator_RW"][0].set_value(value[0])
        self.attribute_mapping["Attenuator_RW"][1].set_value(value[1])
        self.attribute_mapping["Attenuator_RW"][2].set_value(value[2])
        self._Attenuator_RW = value
        # PROTECTED REGION END #    //  RCUSCC.Attenuator_RW_write

    @only_when_on
    @fault_on_opcua_error
    def read_Band_R(self):
        # PROTECTED REGION ID(RCUSCC.Band_R_read) ENABLED START #
        """Return the Band_R attribute."""
        self._Band_R = numpy.array([self.attribute_mapping["Band_R"][0].get_value(), self.attribute_mapping["Band_R"][1].get_value(), self.attribute_mapping["Band_R"][2].get_value()])
        return self._Band_R
        # PROTECTED REGION END #    //  RCUSCC.Band_R_read

    @only_when_on
    @fault_on_opcua_error
    def read_Band_RW(self):
        # PROTECTED REGION ID(RCUSCC.Band_RW_read) ENABLED START #
        """Return the Band_R attribute."""
        self._Band_RW =  numpy.array([self.attribute_mapping["Band_RW"][0].get_value(), self.attribute_mapping["Band_RW"][1].get_value(), self.attribute_mapping["Band_RW"][2].get_value()])
        return self._Band_RW
        # PROTECTED REGION END #    //  RCUSCC.Band_RW_read

    @only_when_on
    @fault_on_opcua_error
    def write_Band_RW(self, value):
        # PROTECTED REGION ID(RCUSCC.Band_RW_write) ENABLED START #
        """Set the Band_RW attribute."""
        self.attribute_mapping["Band_RW"][0].set_value(value[0])
        self.attribute_mapping["Band_RW"][1].set_value(value[1])
        self.attribute_mapping["Band_RW"][2].set_value(value[2])
        self._Band_RW = value
        # PROTECTED REGION END #    //  RCUSCC.Band_RW_write

    @only_when_on
    @fault_on_opcua_error
    def read_Dither_Frequency_R(self):
        # PROTECTED REGION ID(RCUSCC.Dither_Frequency_R_read) ENABLED START #
        """Return the Dither_Frequency_R attribute."""
        self._Dither_Frequency_R = numpy.array([self.attribute_mapping["Dither_Frequency_R"][0].get_value(), self.attribute_mapping["Dither_Frequency_R"][1].get_value()])
        return self._Dither_Frequency_R
        # PROTECTED REGION END #    //  RCUSCC.Dither_Frequency_R_read

    @only_when_on
    @fault_on_opcua_error
    def read_Dither_Frequency_RW(self):
        # PROTECTED REGION ID(RCUSCC.Dither_Frequency_RW_read) ENABLED START #
        """Return the Dither_Frequency_R attribute."""
        self._Dither_Frequency_RW = numpy.array([self.attribute_mapping["Dither_Frequency_RW"][0].get_value(), self.attribute_mapping["Dither_Frequency_RW"][1].get_value()])
        return self._Dither_Frequency_RW
        # PROTECTED REGION END #    //  RCUSCC.Dither_Frequency_RW_read

    @only_when_on
    @fault_on_opcua_error
    def write_Dither_Frequency_RW(self, value):
        # PROTECTED REGION ID(RCUSCC.Dither_Frequency_RW_write) ENABLED START #
        """Set the Dither_Frequency_RW attribute."""
        self.attribute_mapping["Dither_Frequency_RW"][0].set_value(value[0])
        self.attribute_mapping["Dither_Frequency_RW"][1].set_value(value[1])
        self._Dither_Frequency_RW = value
        # PROTECTED REGION END #    //  RCUSCC.Dither_Frequency_RW_write

    @only_when_on
    @fault_on_opcua_error
    def read_LED_R(self):
        # PROTECTED REGION ID(RCUSCC.LED_R_read) ENABLED START #
        """Return the LED_R attribute."""
        self._LED_R = self.attribute_mapping["LED_R"].get_value()
        return self._LED_R
        # PROTECTED REGION END #    //  RCUSCC.LED_R_read

    @only_when_on
    @fault_on_opcua_error
    def read_LED_RW(self):
        # PROTECTED REGION ID(RCUSCC.LED_RW_read) ENABLED START #
        """Return the LED_RW attribute."""
        self._LED_RW = self.attribute_mapping["LED_RW"].get_value()
        return self._LED_RW
        # PROTECTED REGION END #    //  RCUSCC.LED_RW_read

    @only_when_on
    @fault_on_opcua_error
    def write_LED_RW(self, value):
        # PROTECTED REGION ID(RCUSCC.LED_RW_write) ENABLED START #
        """Set the LED_RW attribute."""
        self.attribute_mapping["LED_RW"].set_value(value)
        self._LED_RW = value
        # PROTECTED REGION END #    //  RCUSCC.LED_RW_write

    @only_when_on
    @fault_on_opcua_error
    def read_Pwr_dig_R(self):
        # PROTECTED REGION ID(RCUSCC.Pwr_dig_R_read) ENABLED START #
        """Return the Pwr_dig_R attribute."""
        self._Pwr_dig_R = self.attribute_mapping["Pwr_dig_R"].get_value()
        return self._Pwr_dig_R
        # PROTECTED REGION END #    //  RCUSCC.Pwr_dig_R_read

    @only_when_on
    @fault_on_opcua_error
    def read_Temperature_R(self):
        # PROTECTED REGION ID(RCUSCC.Temperature_R_read) ENABLED START #
        """Return the Temperature_R attribute."""
        self._Temperature_R = self.attribute_mapping["Temperature_R"].get_value()
        return self._Temperature_R
        # PROTECTED REGION END #    //  RCUSCC.Temperature_R_read

    @only_when_on
    @fault_on_opcua_error
    def read_CLK_PLL_locked_R(self):
        # PROTECTED REGION ID(RCUSCC.CLK_PLL_locked_R_read) ENABLED START #
        """Return the CLK_PLL_locked_R attribute."""
        self._CLK_PLL_locked_R = self.attribute_mapping["CLK_PLL_locked_R"].get_value()
        return self._CLK_PLL_locked_R
        # PROTECTED REGION END #    //  RCUSCC.CLK_PLL_locked_R_read


    # --------
    # Commands
    # --------

    @command(
    )
    @DebugIt()
    def On(self):
        # PROTECTED REGION ID(RCUSCC.On) ENABLED START #
        """

        :return:None
        """
        self.set_state(DevState.ON)
        # PROTECTED REGION END #    //  RCUSCC..On

    @command(
    )
    @DebugIt()
    def Off(self):
        # PROTECTED REGION ID(RCUSCC.Off) ENABLED START #
        """

        :return:None
        """
        self.set_state(DevState.OFF)

        # stop reconnecting before disconnect
        self.opcua_connection.stop()

        # PROTECTED REGION END #    //  RCUSCC.Off

    @command(
    )
    @DebugIt()
    def Fault(self):
        # PROTECTED REGION ID(RCUSCC.On) ENABLED START #
        """
        FAULT state is used to indicate our connection with the OPC-UA server is down.

        This device will try to reconnect, and transition to the ON state on success.

        :return:None
        """
        self.set_state(DevState.FAULT)
        # PROTECTED REGION END #    //  RCUSCC..On

    @command(
    )
    @DebugIt()
    def RCU_off(self):
        # PROTECTED REGION ID(RCUSCC.RCU_off) ENABLED START #
        """

        :return:None
        """
#        self.attribute_mapping["RCU_off"]()
        # PROTECTED REGION END #    //  RCUSCC.RCU_off

    @command(
    )
    @DebugIt()
    def RCU_on(self):
        # PROTECTED REGION ID(RCUSCC.RCU_on) ENABLED START #
        """

        :return:None
        """
#        self.attribute_mapping["RCU_on"]()
        # PROTECTED REGION END #    //  RCUSCC.RCU_on


# ----------
# Run server
# ----------


def main(args=None, **kwargs):
    """Main function of the RCUSCC module."""
    # PROTECTED REGION ID(RCUSCC.main) ENABLED START #
    return run((RCUSCC,), args=args, **kwargs)
    # PROTECTED REGION END #    //  RCUSCC.main


if __name__ == '__main__':
    main()
